
-- Copyright (c) 2016 Dr. Charles Mallah
-- http://omicroncubed.com/

--[[
  Permission is hereby granted, free of charge, to any person
  obtaining a copy of this software and associated documentation
  files (the "Software"), to deal in the Software without
  restriction, including without limitation the rights to use,
  copy, modify, merge, publish, distribute, sub-license, and/or sell
  copies of the Software, and to permit persons to whom the
  Software is furnished to do so, subject to the following
  conditions:

  The above copyright notice and this permission notice shall be
  included in all copies or substantial portions of the Software.

  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
  EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
  OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
  NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
  HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
  WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
  OTHER DEALINGS IN THE SOFTWARE.
]]

--- Return a table of full paths of all files in the folder and all sub-folders
-- 
local function recursiveFileSearch(folder, fileTree)
	local lfs        = love.filesystem
	local filesTable = lfs.getDirectoryItems(folder)
  
  if fileTree == nil then
    fileTree = {}
  end
  
	for i = 1, #filesTable do
    local o = filesTable[i]
		local file = folder .. "/" .. o
    
		if lfs.isFile(file) then
      table.insert(fileTree, file)
		elseif lfs.isDirectory(file) then
			recursiveFileSearch(file, fileTree)
		end
	end
  
	return fileTree
end

--- Save a serialised table
--
function tableSaveOut(data, dir, filename)
  if dir ~= "" then
    love.filesystem.createDirectory(dir)
    filename = dir .. filename
  end
  local serialiser = require("serial")
  local serialdata = serialiser:dumps(data)
  love.filesystem.write(filename, serialdata)
end

--- Load a saved serialised table
--
function tableLoadIn(filename)
  if love.filesystem.isFile(filename) then
    local serialiser = require("serial")
    local serialdata = love.filesystem.read(filename)
    do return serialiser:loads(serialdata) end
  end
  return nil
end

--- Repeatedly called from the main function
-- adds an image to a canvas and stores the quad info, or sets quad info into the cache 
local function objectSpritebatchAdd(ATTACHMENT_CANVAS, ATTACHMENT_QUADS, ATLAS_W, 
                                     ATLAS_CACHE_MAPPING, ATLAS_CACHE, rollingAttachmentCacheCount,
                                      source, image, w, h, x, y, currenth)
                              
  if ATLAS_CACHE_MAPPING[source] == nil then
    rollingAttachmentCacheCount          = rollingAttachmentCacheCount + 1
    
    -- Canvas quad
    if ATTACHMENT_CANVAS ~= nil then
      -- Increment y if the current quad will not fit the canvas width
      if x + w > ATLAS_W then
        y = y + currenth + 1
        x = 0
        currenth = h
      end
    end
    
    ATLAS_CACHE_MAPPING[source]              = rollingAttachmentCacheCount -- Store the new index
    ATLAS_CACHE[rollingAttachmentCacheCount] = love.graphics.newQuad(x, y, w, h, ATLAS_W, ATLAS_W) -- ... Cache the quad in the equivalent index
    
    -- Draw image into canvas    
    if ATTACHMENT_CANVAS ~= nil then
      love.graphics.setCanvas(ATTACHMENT_CANVAS)
      
      love.graphics.draw(image, x, y)
      
      love.graphics.setCanvas()
      
      ATTACHMENT_QUADS[rollingAttachmentCacheCount] = { source = source, x = x, y = y, w = w, h = h }
    end
    
    -- Increment x position
    x = x + w + 1
    
    do return rollingAttachmentCacheCount, x, y, currenth end -- Return the index only
  end
  
end

--- Main function to create a new atlas and quad cache
--
function newAtlas(ATLAS_W, path, ATLAS_CACHE_MAPPING, ATLAS_CACHE, spritebatchMaxSprites)
  
  -- Search for all images
  local attachmentSources         = recursiveFileSearch(path)
  local attachmentCount           = #attachmentSources
  
  -- Add to sprite-batch canvas
  local imageSets = {}
  for i = 1, attachmentCount do
    local source  = attachmentSources[i]
    local image   = love.graphics.newImage(source)
    local w, h    = image:getWidth(), image:getHeight()
    imageSets[i]  = {source = source, image = image, w = w, h = h}
  end
  
  
  -- Sort from largest to smallest
  table.sort(imageSets, function(a, b) return a.h > b.h end)
  
  -- 
  local atlasName  = attachmentCount .. "_" .. ATLAS_W .. "_Atlas.png"
  local quadsName  = attachmentCount .. "_" .. ATLAS_W .. "_Quads.lua"
  local atlasFound = love.filesystem.exists(atlasName)
  local image      = nil
  local rollingAttachmentCacheCount = 0
  
  if atlasFound == false then
    -- CREATE A NEW ATLAS + QUADS
    
    -- Add to atlas
    local ATTACHMENT_CANVAS = love.graphics.newCanvas(ATLAS_W, ATLAS_W)
    local ATTACHMENT_QUADS  = {}
    
    -- Add to sprite-batch canvas
    love.graphics.setColor(255, 255, 255, 255)
    
    local x, y, h = 0, 0, imageSets[1].h
    for i = 1, attachmentCount do
      local imageSet = imageSets[i]
      rollingAttachmentCacheCount, x, y, h = objectSpritebatchAdd(ATTACHMENT_CANVAS, ATTACHMENT_QUADS, ATLAS_W, 
                                      ATLAS_CACHE_MAPPING, ATLAS_CACHE, rollingAttachmentCacheCount,
                                      imageSet.source, imageSet.image, imageSet.w, imageSet.h, 
                                      x, y, h)
    end
    
    --
    local data  = ATTACHMENT_CANVAS:newImageData(0, 0, ATLAS_W, ATLAS_W)
    image       = love.graphics.newImage(data)
    
    data:encode("png", atlasName)
    
    tableSaveOut(ATTACHMENT_QUADS, "", quadsName)
    
  else
    
    -- LOAD ATLAS + QUADS IN
    local ATTACHMENT_QUADS = tableLoadIn(quadsName)
    
    for i = 1, attachmentCount do
      local o = ATTACHMENT_QUADS[i]
      objectSpritebatchAdd(ATTACHMENT_CANVAS, nil, ATLAS_W, 
                            ATLAS_CACHE_MAPPING, ATLAS_CACHE, i-1,
                              o.source, nil, o.w, o.h, o.x, o.y, 0)
    end
    
    image = love.graphics.newImage(atlasName)
    
  end
  
  -- Create sprite-batch
  return love.graphics.newSpriteBatch(image, spritebatchMaxSprites, "static")
  
end
